import bz2
import io
import os
import struct
import sys
import time
import libconf

from indexed_gzip import IndexedGzipFile as GzipFile

from cachetools import TTLCache
from hashlib import sha256
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256

from .update_metadata_pb2 import DeltaArchiveManifest
from .update_metadata_pb2 import InstallOperation
from .update_metadata_pb2 import Signatures

from .cpio import Archive


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def range_contains(range1, range2):
    return range1.start < range2.stop and range2.start < range1.stop


class BlockCache(TTLCache):
    def __init__(self, maxsize, ttl, timer=time.monotonic, getsizeof=sys.getsizeof):
        super().__init__(maxsize, ttl, timer, getsizeof)

    @property
    def usage_str(self):
        return f"{self.curr_size_str}/{self.max_size_str}"

    @property
    def curr_size_str(self):
        return sizeof_fmt(self.currsize)

    @property
    def max_size_str(self):
        return sizeof_fmt(self.maxsize)

    def will_fit(self, value) -> bool:
        return self.maxsize >= self.getsizeof(value)


class UpdateImageException(Exception):
    pass


class UpdateImageSignatureException(UpdateImageException):
    def __init__(self, message, signed_hash, actual_hash):
        super().__init__(message)
        self.signed_hash = signed_hash
        self.actual_hash = actual_hash


class ProtobufUpdateImage(io.RawIOBase):
    _manifest = None
    _offset = -1
    _size = 0
    _pos = 0

    def __init__(self, update_file, cache_size=500, cache_ttl=60):
        self.update_file = update_file
        self.cache_size = cache_size
        self._cache = BlockCache(
            maxsize=cache_size * 1024 * 1024,
            ttl=cache_ttl,
        )
        with open(self.update_file, "rb") as f:
            magic = f.read(4)
            if magic != b"CrAU":
                raise UpdateImageException("Wrong header")

            major = struct.unpack(">Q", f.read(8))[0]
            if major != 1:
                raise UpdateImageException("Unsupported version")

            size = struct.unpack(">Q", f.read(8))[0]
            data = f.read(size)
            self._manifest = DeltaArchiveManifest.FromString(data)
            self._offset = f.tell()

        for blob, offset, length, f in self._blobs:
            self._size += length

    def verify(self, publickey):
        _publickey = load_pem_public_key(publickey)
        with open(self.update_file, "rb") as f:
            data = f.read(self._offset + self._manifest.signatures_offset)

        actual_hash = sha256(data).digest()
        signed_hash = _publickey.recover_data_from_signature(
            self.signature,
            PKCS1v15(),
            SHA256(),
        )
        if actual_hash != signed_hash:
            raise UpdateImageSignatureException(
                "Actual hash does not match signed hash", signed_hash, actual_hash
            )

    @property
    def block_size(self):
        return self._manifest.block_size

    @property
    def signature(self):
        for signature in self._signatures:
            if signature.version == 2:
                return signature.data

        return None

    @property
    def _signatures(self):
        with open(self.update_file, "rb") as f:
            f.seek(self._offset + self._manifest.signatures_offset)
            for signature in Signatures.FromString(
                f.read(self._manifest.signatures_size)
            ).signatures:
                yield signature

    @property
    def _blobs(self):
        with open(self.update_file, "rb") as f:
            for blob in self._manifest.partition_operations:
                f.seek(self._offset + blob.data_offset)
                dst_offset = blob.dst_extents[0].start_block * self.block_size
                dst_length = blob.dst_extents[0].num_blocks * self.block_size
                if blob.type not in (0, 1):
                    raise UpdateImageException(f"Unsupported type {blob.type}")

                yield blob, dst_offset, dst_length, f

        self.expire()

    def _read_blob(self, blob, blob_offset, blob_length, f):
        if blob_offset in self._cache:
            return self._cache[blob_offset]

        if blob.type not in (
            InstallOperation.Type.REPLACE,
            InstallOperation.Type.REPLACE_BZ,
        ):
            raise NotImplementedError(
                f"Error: {InstallOperation.Type.keys()[blob.type]} has not been implemented yet"
            )

        blob_data = f.read(blob.data_length)
        if sha256(blob_data).digest() != blob.data_sha256_hash:
            raise UpdateImageException("Error: Data has wrong sha256sum")

        if blob.type == InstallOperation.Type.REPLACE_BZ:
            try:
                blob_data = bz2.decompress(blob_data)

            except ValueError as err:
                raise UpdateImageException(f"Error: {err}") from err

            if blob_length - len(blob_data) < 0:
                raise UpdateImageException(
                    f"Error: Bz2 compressed data was too large {len(blob_data)}"
                )

        # Zero padd data to fit
        if len(blob_data) < blob_length:
            blob_data += b"\0" * (blob_length - len(blob_data))

        assert len(blob_data) == blob_length
        if self._cache.will_fit(blob_data):
            self._cache[blob_offset] = blob_data

        return blob_data

    @property
    def cache(self):
        return self._cache

    @property
    def size(self):
        return self._size

    def expire(self):
        self._cache.expire()

    def writable(self):
        return False

    def seekable(self):
        return True

    def readable(self):
        return True

    def seek(self, offset, whence=os.SEEK_SET):
        if whence not in (os.SEEK_SET, os.SEEK_CUR, os.SEEK_END):
            raise OSError("Not supported whence")
        if whence == os.SEEK_SET and offset < 0:
            raise ValueError("offset can't be negative")
        if whence == os.SEEK_END and offset > 0:
            raise ValueError("offset can't be positive")

        if whence == os.SEEK_SET:
            self._pos = min(max(offset, 0), self._size)
        elif whence == os.SEEK_CUR:
            self._pos = min(max(self._pos + offset, 0), self._size)
        elif whence == os.SEEK_END:
            self._pos = min(max(self._size + offset, 0), self._size)
        return self._pos

    def tell(self):
        return self._pos

    def read(self, size=-1):
        res = self.peek(size)
        self.seek(len(res), whence=os.SEEK_CUR)
        return res

    def peek(self, size=0):
        offset = self._pos
        if offset >= self._size:
            return b""

        if size <= 0 or offset + size > self._size:
            size = self._size - offset

        res = bytearray(size)
        for blob, blob_offset, blob_length, f in self._blobs:
            if not range_contains(
                range(offset, offset + size),
                range(blob_offset, blob_offset + blob_length),
            ):
                if size >= self._size:
                    print(f"Skipping blob {blob_offset} to {blob_length}, {offset}")

                continue

            blob_data = self._read_blob(blob, blob_offset, blob_length, f)
            blob_start_offset = max(offset - blob_offset, 0)
            blob_end_offset = min(offset - blob_offset + size, blob_length)
            data = blob_data[blob_start_offset:blob_end_offset]

            assert blob_start_offset >= 0, (
                f"blob start offset is negative number: {blob_start_offset}"
            )
            assert blob_end_offset <= blob_length, (
                f"blob end offset is larger than blob length: {blob_end_offset}, {blob_length}"
            )
            assert blob_end_offset - blob_start_offset == len(data), (
                f"blob start and end is larger than data: {blob_end_offset - blob_start_offset}, {len(data)}"
                + f"\n  offset: {offset}"
                + f"\n  blob_offset: {blob_offset}"
                + f"\n  size: {size}"
                + f"\n  blob_length: {blob_length}"
                + f"\n  blob_start_offset: {blob_start_offset}"
                + f"\n  blob_end_offset: {blob_end_offset}"
                + f"\n  len(blob_data): {len(blob_data)}"
                + f"\n  blob.type: {blob.type}"
            )

            start_offset = blob_offset + blob_start_offset - offset
            end_offset = blob_offset + blob_end_offset - offset
            res[start_offset:end_offset] = data

            assert start_offset >= 0, f"start offset is negative number: {start_offset}"
            assert start_offset < len(res), (
                f"start offset is larger than size of data: {start_offset}, {len(res)}"
            )
            assert end_offset <= blob_offset + blob_length, (
                f"end offset is larger than size of blob: {end_offset}, {blob_offset + blob_length}"
            )
            assert end_offset - start_offset == len(data), (
                f"size of offsets does not equal size of data, {end_offset - start_offset}, {len(data)}"
            )
            assert end_offset <= len(res), (
                f"end offset is larger than size of data, {end_offset}, {len(res)}"
            )
            assert res[start_offset:end_offset] == data, "data does not match"

        assert len(res) == size, (
            f"size of data does not match expected size: {len(res)}, {size}"
        )
        return bytes(res)


class CPIOUpdateImage(io.RawIOBase):
    def __init__(self, update_file, cache_size=500, cache_ttl=60):
        self.update_file = update_file
        self.cache_size = cache_size
        self._cache = BlockCache(
            maxsize=cache_size * 1024 * 1024,
            ttl=cache_ttl,
        )
        self._archive = Archive(self.update_file)
        self._archive.open()
        if b"sw-description" not in self._archive.keys():
            raise UpdateImageException("Not a swupdate file")

        info = libconf.loads(self._archive["sw-description"].read().decode("utf-8"))[
            "software"
        ]
        self._version = info.get("version")

        if "reMarkable1" in info:
            self._hardware_type = "reMarkable1"
            self._info = info["reMarkable1"]

        elif "reMarkable2" in info:
            self._hardware_type = "reMarkable2"
            self._info = info["reMarkable2"]

        elif "ferrari" in info:
            self._hardware_type = "ferrari"
            self._info = info["ferrari"]

        elif "chiappa" in info:
            self._hardware_type = "chiappa"
            self._info = info["chiappa"]

        else:
            raise UpdateImageException("Unsupported swupdate file")

        # TODO - handle non-stable images
        # TODO - handle possibilities of multiple images
        info = self._info["stable"]["copy1"]["images"][0]
        entry = self._archive[info["filename"]]
        self._image = GzipFile(fileobj=entry, mode="rb")

    def verify(self, publickey):
        # TODO - verify signature
        def verify_hash(expected_hash, entry):
            actual_hash = sha256(entry.peek()).hexdigest()
            if expected_hash != actual_hash:
                raise UpdateImageException(
                    "Actual hash does not match metadata hash",
                    expected_hash,
                    actual_hash,
                )

        def verify_copy(copy):
            for image in copy["images"]:
                verify_hash(image["sha256"], self._archive[image["filename"]])

            for image in copy.get("files", []):
                verify_hash(image["sha256"], self._archive[image["filename"]])

            for image in copy.get("scripts", []):
                verify_hash(image["sha256"], self._archive[image["filename"]])

        for copy in ("copy1", "copy2"):
            verify_copy(self._info["stable"][copy])

    @property
    def signature(self):
        # TODO - get from entry
        return None

    @property
    def version(self) -> str | None:
        return self._version

    @property
    def hardware_type(self) -> str:
        return self._hardware_type

    @property
    def archive(self):
        return self._archive

    @property
    def cache(self):
        return self._cache

    @property
    def size(self):
        return self._image.size

    def expire(self):
        self._cache.expire()

    def close(self):
        try:
            self._archive.close()

        finally:
            super().close()

    def writable(self):
        return False

    def seekable(self):
        return True

    def readable(self):
        return True

    def seek(self, offset, whence=os.SEEK_SET):
        return self._image.seek(offset, whence)

    def tell(self):
        return self._image.tell()

    def read(self, size=-1):
        key = (self.tell(), size)
        if key in self._cache:
            return self._cache[key]

        data = self._image.read(size)
        if self._cache.will_fit(data):
            self._cache[key] = data

        return data

    def peek(self, size=0):
        key = (self.tell(), size)
        if key in self._cache:
            return self._cache[key]

        data = self._image.peek(size)
        if self._cache.will_fit(data):
            self._cache[key] = data

        return data


class UpdateImage:
    def __new__(cls, *args, **kwds):
        try:
            return ProtobufUpdateImage(*args, **kwds)

        except UpdateImageException:
            pass

        return CPIOUpdateImage(*args, **kwds)



import io
import errno

from struct import pack

from ctypes import c_ushort
from ctypes import c_char
from ctypes import sizeof
from ctypes import Structure
from ctypes import LittleEndianStructure
from ctypes import BigEndianStructure


class MagicError(Exception):
    pass


class ChecksumError(Exception):
    pass


def to_hex(data):
    if isinstance(data, int):
        return f"0x{data:02X}"

    return "0x" + "".join([f"{x:02X}" for x in data])


_header_old_cpio = [
    ("c_magic", c_ushort),
    ("c_dev", c_ushort),
    ("c_ino", c_ushort),
    ("c_mode", c_ushort),
    ("c_uid", c_ushort),
    ("c_gid", c_ushort),
    ("c_nlink", c_ushort),
    ("c_rdev", c_ushort),
    ("c_mtime", c_ushort * 2),
    ("c_namesize", c_ushort),
    ("c_filesize", c_ushort * 2),
]


class header_old_cpio_le(LittleEndianStructure):
    _fields_ = _header_old_cpio


class header_old_cpio_be(BigEndianStructure):
    _fields_ = _header_old_cpio


def _header_old_cpio_verify(self) -> None:
    if self.c_magic != 0o070707:
        raise MagicError(
            f"{self} magic bytes do not match! "
            f"expected={to_hex(0o070707)}, "
            f"actual={to_hex(self.c_magic)}"
        )


def _header_old_cpio_namesize(self) -> int:
    return self.c_namesize.value


def _header_old_cpio_filesize(self) -> int:
    return self.c_filesize.value


def _header_old_cpio_entrysize(self) -> int:
    return self.size + self.filesize + (self.filesize % 2)


def _header_old_cpio_size(self) -> int:
    return sizeof(self) + self.namesize + (self.namesize % 2)


header_old_cpio_le.verify = _header_old_cpio_verify
header_old_cpio_be.verify = _header_old_cpio_verify
header_old_cpio_le.namesize = property(_header_old_cpio_namesize)
header_old_cpio_be.namesize = property(_header_old_cpio_namesize)
header_old_cpio_le.filesize = property(_header_old_cpio_filesize)
header_old_cpio_be.filesize = property(_header_old_cpio_filesize)
header_old_cpio_le.entrysize = property(_header_old_cpio_entrysize)
header_old_cpio_be.entrysize = property(_header_old_cpio_entrysize)
header_old_cpio_le.size = property(_header_old_cpio_size)
header_old_cpio_be.size = property(_header_old_cpio_size)

_cpio_odc_header = [
    ("c_magic", c_char * 6),
    ("c_dev", c_char * 6),
    ("c_ino", c_char * 6),
    ("c_mode", c_char * 6),
    ("c_uid", c_char * 6),
    ("c_gid", c_char * 6),
    ("c_nlink", c_char * 6),
    ("c_rdev", c_char * 6),
    ("c_mtime", c_char * 11),
    ("c_namesize", c_char * 6),
    ("c_filesize", c_char * 11),
]


class cpio_odc_header_le(LittleEndianStructure):
    _fields_ = _cpio_odc_header


class cpio_odc_header_be(BigEndianStructure):
    _fields_ = _cpio_odc_header


def _cpio_odc_header_verify(self) -> None:
    if self.c_magic != "070707":
        raise MagicError(
            f"{self} magic bytes do not match! "
            f"expected={'070707'}, "
            f"actual={self.c_magic}"
        )


def _cpio_odc_header_namesize(self) -> int:
    return int(self.c_namesize, 8)


def _cpio_odc_header_filesize(self) -> int:
    return int(self.c_filesize, 8)


def _cpio_odc_header_entrysize(self) -> int:
    return self.size + self.filesize + (self.filesize % 2)


def _cpio_odc_header_size(self) -> int:
    return sizeof(self) + self.namesize + (self.namesize % 2)


cpio_odc_header_le.verify = _cpio_odc_header_verify
cpio_odc_header_be.verify = _cpio_odc_header_verify
cpio_odc_header_le.namesize = property(_cpio_odc_header_namesize)
cpio_odc_header_be.namesize = property(_cpio_odc_header_namesize)
cpio_odc_header_le.filesize = property(_cpio_odc_header_filesize)
cpio_odc_header_be.filesize = property(_cpio_odc_header_filesize)
cpio_odc_header_le.entrysize = property(_cpio_odc_header_entrysize)
cpio_odc_header_be.entrysize = property(_cpio_odc_header_entrysize)
cpio_odc_header_le.size = property(_cpio_odc_header_size)
cpio_odc_header_be.size = property(_cpio_odc_header_size)

_cpio_newc_header = [
    ("c_magic", c_char * 6),
    ("c_ino", c_char * 8),
    ("c_mode", c_char * 8),
    ("c_uid", c_char * 8),
    ("c_gid", c_char * 8),
    ("c_nlink", c_char * 8),
    ("c_mtime", c_char * 8),
    ("c_filesize", c_char * 8),
    ("c_devmajor", c_char * 8),
    ("c_devminor", c_char * 8),
    ("c_rdevmajor", c_char * 8),
    ("c_rdevminor", c_char * 8),
    ("c_namesize", c_char * 8),
    ("c_check", c_char * 8),
]


class cpio_newc_header_le(LittleEndianStructure):
    _fields_ = _cpio_newc_header


class cpio_newc_header_be(BigEndianStructure):
    _fields_ = _cpio_newc_header


def _cpio_newc_header_verify(self) -> None:
    if self.c_magic not in (b"070701", b"070702"):
        raise MagicError(
            f"{self} magic bytes do not match! "
            f"expected=070701 or 070702, "
            f"actual={self.c_magic}"
        )

    if self.c_magic == b"070702":
        # TODO - verify data
        pass


def _cpio_newc_header_namesize(self) -> int:
    return int(self.c_namesize, 16)


def _cpio_newc_header_filesize(self) -> int:
    return int(self.c_filesize, 16)


def _cpio_newc_header_entrysize(self) -> int:
    size = self.size + self.filesize
    if self.filesize % 4:
        size += 4 - (self.filesize % 4)

    return size


def _cpio_newc_header_size(self) -> int:
    size = sizeof(self) + self.namesize
    if size % 4:
        size += 4 - (size % 4)

    return size


cpio_newc_header_le.verify = _cpio_newc_header_verify
cpio_newc_header_be.verify = _cpio_newc_header_verify
cpio_newc_header_le.namesize = property(_cpio_newc_header_namesize)
cpio_newc_header_be.namesize = property(_cpio_newc_header_namesize)
cpio_newc_header_le.filesize = property(_cpio_newc_header_filesize)
cpio_newc_header_be.filesize = property(_cpio_newc_header_filesize)
cpio_newc_header_le.entrysize = property(_cpio_newc_header_entrysize)
cpio_newc_header_be.entrysize = property(_cpio_newc_header_entrysize)
cpio_newc_header_le.size = property(_cpio_newc_header_size)
cpio_newc_header_be.size = property(_cpio_newc_header_size)


class Entry:
    def __init__(self, fileobj, offset):
        self.fileobj = fileobj
        self.offset = offset
        self.cursor = 0
        self.header = self.read_header()
        self.header.verify()
        self.dataoffset = offset + self.header.size

    def __len__(self):
        return self.header.filesize

    def read_header(self) -> Structure:
        self.fileobj.seek(self.offset)
        magic = self.fileobj.read(sizeof(c_ushort))
        if magic == pack(">H", 0o070707):
            cls = header_old_cpio_be

        elif magic == pack("<H", 0o070707):
            cls = header_old_cpio_le

        else:
            self.fileobj.seek(self.offset)
            magic = self.fileobj.read(sizeof(c_char * 6))
            if magic == b"070707":
                cls = cpio_odc_header_le

            # elif magic == b"070707":
            #     cls = cpio_odc_header_be

            elif magic in (b"070701", b"070702"):
                cls = cpio_newc_header_le

            # elif magic in (b"070701", b"070702"):
            #     cls = cpio_newc_header_be

            else:
                raise MagicError(f"Unknown magic: {magic}")

        self.fileobj.seek(self.offset)
        data = self.fileobj.read(sizeof(cls))
        return cls.from_buffer_copy(data)

    @property
    def name(self) -> bytes:
        self.fileobj.seek(self.offset + sizeof(self.header))
        return self.fileobj.read(self.header.namesize).rstrip(b"\x00")

    @property
    def data(self) -> bytes:
        self.fileobj.seek(self.dataoffset)
        return self.fileobj.read(self.header.filesize)

    @property
    def size(self):
        return len(self)

    def writable(self):
        return False

    def seekable(self):
        return True

    def readable(self):
        return True

    def seek(self, offset, mode=io.SEEK_SET) -> None:
        if mode == io.SEEK_CUR:
            offset += self.cursor

        elif mode == io.SEEK_END:
            offset += len(self)

        elif mode != io.SEEK_SET:
            raise NotImplementedError()

        if offset < 0:
            raise OSError(errno.EINVAL, "Invalid argument")

        self.cursor = offset

    def tell(self) -> int:
        return self.cursor

    def read(self, size=-1) -> bytes:
        if size < 0:
            size = len(self) - self.cursor

        data = self.peek(size)
        self.cursor += len(data)
        if size < len(data):
            raise OSError(errno.EIO, "Unexpected EOF")

        return data

    def peek(self, size=0) -> bytes:
        if self.cursor >= len(self):
            return b""

        if not size or size + self.cursor > len(self):
            size = len(self) - self.cursor

        self.fileobj.seek(self.dataoffset + self.cursor)
        return self.fileobj.read(size)


class Archive:
    def __init__(self, fileOrPath):
        self.fileOrPath = fileOrPath
        self.fileobj = None
        self.entries = {}

    def __enter__(self):
        self.open()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __getitem__(self, key: str | bytes) -> Entry | None:
        return self.get(key)

    def __len__(self) -> int:
        return len(self.entries)

    def open(self):
        if self.fileobj is not None:
            return

        if isinstance(self.fileOrPath, str):
            self.fileobj = open(self.fileOrPath, "rb")

        else:
            self.fileobj = self.fileOrPath

        offset = 0
        self.fileobj.seek(0, io.SEEK_END)
        size = self.fileobj.tell()
        self.fileobj.seek(0, io.SEEK_SET)
        while offset < size:
            entry = Entry(self.fileobj, offset)
            if entry.name == b"TRAILER!!!":
                break

            self.entries[entry.name] = entry
            offset += entry.header.entrysize

    def close(self):
        if isinstance(self.fileOrPath, str):
            self.fileobj.close()

        self.fileobj = None
        self.entries = {}

    def get(self, name: str | bytes, default=None) -> Entry | None:
        if isinstance(name, str):
            name = name.encode("ascii")

        if not isinstance(name, bytes):
            raise NotImplementedError()

        return self.entries.get(name, default)

    def keys(self) -> list[bytes]:
        return self.entries.keys()

    def values(self) -> list[Entry]:
        return self.entries.values()
